import io
import json
from typing import Any, Callable, Dict, List, Optional, Tuple

import pandas as pd
from aqueduct.utils.format import DEFAULT_ENCODING
from aqueduct.utils.serialization import (
    artifact_type_to_serialization_type,
    deserialize,
    serialize_val,
)
from aqueduct_executor.operators.utils.enums import (
    ArtifactType,
    ExecutionStatus,
    FailureType,
    SerializationType,
)
from aqueduct_executor.operators.utils.exceptions import MissingInputPathsException
from aqueduct_executor.operators.utils.execution import (
    TIP_UNKNOWN_ERROR,
    Error,
    ExecFailureException,
    ExecutionState,
    Logs,
    exception_traceback,
)
from aqueduct_executor.operators.utils.saved_object_delete import SavedObjectDelete
from aqueduct_executor.operators.utils.storage.storage import Storage
from aqueduct_executor.operators.utils.utils import (
    _METADATA_ARTIFACT_TYPE_KEY,
    _METADATA_PYTHON_TYPE_KEY,
    _METADATA_SCHEMA_KEY,
    _METADATA_SERIALIZATION_TYPE_KEY,
    _METADATA_SYSTEM_METADATA_KEY,
    serialize_val_wrapper,
)
from pyspark.sql import SparkSession


def read_artifacts_spark(
    storage: Storage,
    input_paths: List[str],
    input_metadata_paths: List[str],
    spark_session_obj: SparkSession,
) -> Tuple[List[Any], List[ArtifactType], List[SerializationType]]:
    if len(input_paths) != len(input_metadata_paths):
        raise Exception(
            "Found inconsistent number of input paths (%d) and input metadata paths (%d)"
            % (
                len(input_paths),
                len(input_metadata_paths),
            )
        )

    inputs: List[Any] = []
    artifact_types: List[ArtifactType] = []
    serialization_types: List[SerializationType] = []

    for input_path, input_metadata_path in zip(input_paths, input_metadata_paths):
        artifact_metadata = json.loads(storage.get(input_metadata_path).decode(DEFAULT_ENCODING))
        artifact_type = artifact_metadata[_METADATA_ARTIFACT_TYPE_KEY]
        artifact_types.append(artifact_type)

        serialization_type = artifact_metadata[_METADATA_SERIALIZATION_TYPE_KEY]
        serialization_types.append(serialization_type)

        # Check if artifact is of type TABLE. If it is, attempt to read from a temporary view with
        # name of the input_path.
        if artifact_type != ArtifactType.TABLE:
            inputs.append(deserialize(serialization_type, artifact_type, storage.get(input_path)))
        else:
            # read from temp view
            try:
                # global_temp_db = spark_session_obj.conf.get("spark.sql.globalTempDatabase")
                view_path = "global_temp" + "." + convert_path_to_view_name(input_path)
                spark_df = spark_session_obj.read.table(view_path)
                inputs.append(spark_df)
            except Exception as e:
                raise MissingInputPathsException(
                    "Unable to read inputs artifacts from temp view. Exception: %s" % str(e)
                )

    return inputs, artifact_types, serialization_types


def write_artifact_spark(
    storage: Storage,
    artifact_type: ArtifactType,
    # derived_from_bson specifies if the artifact is derived from a bson object
    # and thus requires bson encoding.
    # For now, it only applies to data frames extracted / transformed from Mongo.
    derived_from_bson: bool,
    output_path: Optional[str],
    output_metadata_path: str,
    content: Any,
    system_metadata: Dict[str, str],
    spark_session_obj: SparkSession,
) -> None:
    """The `output_path` can be empty if the contents were already pre-populated (eg. parameter operators)."""
    output_metadata: Dict[str, Any] = {
        _METADATA_SCHEMA_KEY: [],
        _METADATA_SYSTEM_METADATA_KEY: system_metadata,
        _METADATA_ARTIFACT_TYPE_KEY: artifact_type.value,
    }

    if artifact_type == ArtifactType.TABLE:
        output_metadata[_METADATA_SCHEMA_KEY] = [{col[0]: col[1]} for col in content.dtypes]

    serialization_type = artifact_type_to_serialization_type(
        artifact_type, derived_from_bson, content
    ).value

    if output_path is not None:
        if artifact_type == ArtifactType.TABLE:
            # write artifact to temp view
            # take a sample of the DF
            # write that to aqueduct storage
            spark_df = content
            global_view_name = convert_path_to_view_name(output_path)
            spark_df.createOrReplaceGlobalTempView(global_view_name)
            pandas_df = spark_df.limit(100).toPandas()
            serialized_val = serialize_val_wrapper(pandas_df, serialization_type, derived_from_bson)
            storage.put(output_path, serialized_val)

        else:
            serialized_val = serialize_val_wrapper(content, serialization_type, derived_from_bson)
            storage.put(output_path, serialized_val)

    output_metadata[_METADATA_SERIALIZATION_TYPE_KEY] = serialization_type
    output_metadata[_METADATA_PYTHON_TYPE_KEY] = type(content).__name__
    storage.put(output_metadata_path, json.dumps(output_metadata).encode(DEFAULT_ENCODING))


def convert_path_to_view_name(path: str) -> str:
    """
    Converts a given input/output path for an artifact and converts
    it into a Spark Temporary View compatible name. THis will convert
    slashes and hyphens into underscores.
    """
    return path.replace("-", "_").replace("/", "_")



from typing import Any, Callable, Dict, List, Optional

from aqueduct_executor.operators.connectors.data import (
    common,
    config,
    connector,
    extract,
    load,
    relational,
    snowflake,
    utils,
)
from aqueduct_executor.operators.utils.enums import ArtifactType
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.functions import col
from pyspark.sql.types import FloatType


class SparkSnowflakeConnector(relational.RelationalConnector):
    def __init__(self, config: config.SnowflakeConfig):
        url = "https://{account_identifier}.snowflakecomputing.com".format(
            account_identifier=config.account_identifier,
        )
        conn_engine = snowflake.create_snowflake_engine(config)

        self.snowflake_spark_options = {
            "sfURL": url,
            "sfAccount": config.account_identifier,
            "sfUser": config.username,
            "sfPassword": config.password,
            "sfDatabase": config.database,
            "sfSchema": config.schema,
            "sfWarehouse": config.warehouse,
        }
        super().__init__(conn_engine)

    def extract_spark(
        self, params: extract.RelationalParams, spark_session_obj: SparkSession
    ) -> Any:
        assert params.usable(), "Query is not usable. Did you forget to expand placeholders?"

        df = (
            spark_session_obj.read.format("snowflake")
            .options(**self.snowflake_spark_options)
            .option("query", params.query)
            .load()
        )
        decimals_cols = [c for c in df.columns if "Decimal" in str(df.schema[c].dataType)]
        # convert all decimals columns to floats
        for col in decimals_cols:
            df = df.withColumn(col, df[col].cast(FloatType()))
        return df

    def load_spark(
        self, params: load.RelationalParams, df: DataFrame, artifact_type: ArtifactType
    ) -> None:
        snowflake_update_mode = _map_to_snowflake_mode(params)

        df.write.format("snowflake").options(**self.snowflake_spark_options).option(
            "sfSchema", "public"
        ).option("dbtable", params.table).mode(snowflake_update_mode).save()


def _map_to_snowflake_mode(params: load.RelationalParams) -> str:
    """
    Map Aqueduct update modes to Snowflake update modes found here:
    https://docs.snowflake.com/ko/developer-guide/snowpark/reference/python/api/snowflake.snowpark.DataFrameWriter.save_as_table.html
    """
    snowflake_update_mode = "ignore"
    update_mode = params.update_mode
    if update_mode == common.UpdateMode.APPEND:
        snowflake_update_mode = "append"
    elif update_mode == common.UpdateMode.REPLACE:
        snowflake_update_mode = "overwrite"
    elif update_mode == common.UpdateMode.FAIL:
        snowflake_update_mode = "errorifexists"

    return snowflake_update_mode



import importlib
import json
import os
import shutil
import sys
import tracemalloc
import uuid
from typing import Any, Callable, Dict, List, Tuple

import numpy as np
import pandas as pd
from aqueduct.utils.type_inference import infer_artifact_type
from aqueduct_executor.operators.function_executor import extract_function, get_extract_path
from aqueduct_executor.operators.function_executor.execute import (
    cleanup,
    get_py_import_path,
    import_invoke_method,
    validate_spec,
)
from aqueduct_executor.operators.function_executor.spec import FunctionSpec
from aqueduct_executor.operators.function_executor.utils import OP_DIR
from aqueduct_executor.operators.spark.utils import read_artifacts_spark, write_artifact_spark
from aqueduct_executor.operators.utils import utils
from aqueduct_executor.operators.utils.enums import (
    ArtifactType,
    CheckSeverity,
    ExecutionStatus,
    FailureType,
    OperatorType,
    SerializationType,
)
from aqueduct_executor.operators.utils.execution import (
    TIP_CHECK_DID_NOT_PASS,
    TIP_NOT_BOOL,
    TIP_NOT_NUMERIC,
    TIP_OP_EXECUTION,
    TIP_UNKNOWN_ERROR,
    ExecFailureException,
    ExecutionState,
    Logs,
    exception_traceback,
)
from aqueduct_executor.operators.utils.storage.parse import parse_storage
from aqueduct_executor.operators.utils.timer import Timer
from pyspark.sql import SparkSession, dataframe


def _infer_artifact_type_spark(value: Any) -> Any:
    if isinstance(value, dataframe.DataFrame):
        return ArtifactType.TABLE
    else:
        return infer_artifact_type(value)


def _execute_function(
    spec: FunctionSpec,
    inputs: List[Any],
    exec_state: ExecutionState,
) -> Tuple[List[Any], Dict[str, str]]:
    """
    Invokes the given function on the input data. Does not raise an exception on any
    user function errors, but instead annotates the given exec state with the error.

    :param inputs: the input data to feed into the user's function.
    """

    invoke = import_invoke_method(spec)
    timer = Timer()
    timer.start()
    tracemalloc.start()

    @exec_state.user_fn_redirected(failure_tip=TIP_OP_EXECUTION)
    def _invoke() -> Any:
        return invoke(*inputs)

    results = _invoke()
    if len(spec.output_content_paths) == 1:
        results = [results]

    elapsedTime = timer.stop()
    _, peak = tracemalloc.get_traced_memory()
    system_metadata = {
        utils._RUNTIME_SEC_METRIC_NAME: str(elapsedTime),
        utils._MAX_MEMORY_MB_METRIC_NAME: str(peak / 10**6),
    }

    sys.path.pop(0)
    return results, system_metadata


def _validate_result_count_and_infer_type(
    spec: FunctionSpec,
    results: List[Any],
) -> List[ArtifactType]:
    """
    Validates that the expected number of results were returned by the Function
    and infers the ArtifactType of each result.

    Args:
        spec: The FunctionSpec for the Function
        results: The results returned by the Function

    Returns:
        The ArtifactType of each result

    Raises:
        ExecFailureException: If the expected number of results were not returned
    """
    if len(spec.output_content_paths) > 1 and len(spec.output_content_paths) != len(results):
        raise ExecFailureException(
            failure_type=FailureType.USER_FATAL,
            tip="Expected function to have %s outputs, but instead it had %s."
            % (len(spec.output_content_paths), len(results)),
        )

    return [_infer_artifact_type_spark(res) for res in results]


def run(spec: FunctionSpec, spark_session_obj: SparkSession) -> None:
    """
    Executes a function operator.
    """
    print("Started %s job: %s" % (spec.type, spec.name))

    exec_state = ExecutionState(user_logs=Logs())
    storage = parse_storage(spec.storage_config)
    try:
        validate_spec(spec)

        # Read the input data from intermediate storage.
        inputs, _, serialization_types = read_artifacts_spark(
            storage,
            spec.input_content_paths,
            spec.input_metadata_paths,
            spark_session_obj,
        )

        derived_from_bson = SerializationType.BSON_TABLE in serialization_types
        print("Invoking the function...")
        results, system_metadata = _execute_function(spec, inputs, exec_state)
        if exec_state.status == ExecutionStatus.FAILED:
            # user failure
            utils.write_exec_state(storage, spec.metadata_path, exec_state)
            sys.exit(1)

        print("Function invoked successfully!")

        result_types = _validate_result_count_and_infer_type(spec, results)

        # Perform type checking on the function output.
        if spec.operator_type == OperatorType.METRIC:
            assert len(results) == 1, "Metric operator can only have a single output."
            result = results[0]

            if not (
                isinstance(result, int)
                or isinstance(result, float)
                or isinstance(result, np.number)
            ):
                raise ExecFailureException(
                    failure_type=FailureType.USER_FATAL,
                    tip=TIP_NOT_NUMERIC,
                )

        elif spec.operator_type == OperatorType.CHECK:
            assert len(results) == 1, "Check operator can only have a single output."
            check_result = results[0]

            if isinstance(check_result, pd.Series) and check_result.dtype == "bool":
                assert result_types[0] == ArtifactType.PICKLABLE

                # Cast pd.Series to a bool.
                # We only write True if every boolean in the series is True.
                series = pd.Series(check_result)
                check_passed = bool(series.size - series.sum().item() == 0)
            elif isinstance(check_result, bool) or isinstance(check_result, np.bool_):
                # Cast np.bool_ to a bool.
                check_passed = bool(check_result)
            else:
                raise ExecFailureException(
                    failure_type=FailureType.USER_FATAL,
                    tip=TIP_NOT_BOOL,
                )

            # If the check returned a value we interpret to mean 'false', we exit here, but
            # not before recording the output artifact value (which will be False).
            if not check_passed:
                print(f"Check Operator did not pass.")

                write_artifact_spark(
                    storage,
                    ArtifactType.BOOL,
                    derived_from_bson,  # derived_from_bson doesn't apply to bool artifact
                    spec.output_content_paths[0],
                    spec.output_metadata_paths[0],
                    check_passed,
                    system_metadata=system_metadata,
                    spark_session_obj=spark_session_obj,
                )

                check_severity = spec.check_severity
                if spec.check_severity is None:
                    print(
                        "Check operator has an unspecified severity on spec. Defaulting to ERROR."
                    )
                    check_severity = CheckSeverity.ERROR

                failure_type = FailureType.USER_FATAL
                if check_severity == CheckSeverity.WARNING:
                    failure_type = FailureType.USER_NON_FATAL

                raise ExecFailureException(failure_type, tip=TIP_CHECK_DID_NOT_PASS)

            # If we get here, we know that the check has passed. The artifact type might need
            # still be updated. Eg. if the output was a pandas series.
            result_types[0] = ArtifactType.BOOL
            results[0] = True
        else:
            for i, expected_output_type in enumerate(spec.expected_output_artifact_types):
                if (
                    expected_output_type != ArtifactType.UNTYPED
                    and expected_output_type != result_types[i]
                ):
                    raise ExecFailureException(
                        failure_type=FailureType.USER_FATAL,
                        tip="Expected type %s for the %d-th output of function, but it is of type %s."
                        % (expected_output_type, i, result_types[i]),
                    )

        for i, result in enumerate(results):
            write_artifact_spark(
                storage,
                result_types[i],
                derived_from_bson,
                spec.output_content_paths[i],
                spec.output_metadata_paths[i],
                result,
                system_metadata=system_metadata,
                spark_session_obj=spark_session_obj,
            )

        # If we made it here, then the operator has succeeded.
        exec_state.status = ExecutionStatus.SUCCEEDED
        print(f"Succeeded! Full logs: {exec_state.json()}")
        utils.write_exec_state(storage, spec.metadata_path, exec_state)

    except ExecFailureException as e:
        # We must reconcile the user logs here, since those logs are not captured on the exception.
        from_exception_exec_state = ExecutionState.from_exception(e, user_logs=exec_state.user_logs)
        print(f"Failed with error. Full Logs:\n{from_exception_exec_state.json()}")
        utils.write_exec_state(storage, spec.metadata_path, from_exception_exec_state)
        sys.exit(1)

    except Exception as e:
        exec_state.mark_as_failure(
            FailureType.SYSTEM, TIP_UNKNOWN_ERROR, context=exception_traceback(e)
        )
        print(f"Failed with system error. Full Logs:\n{exec_state.json()}")
        utils.write_exec_state(storage, spec.metadata_path, exec_state)
        sys.exit(1)
    finally:
        # Perform any cleanup
        cleanup(spec)



import argparse
import base64

from aqueduct_executor.operators.function_executor import execute
from aqueduct_executor.operators.function_executor.spec import parse_spec
from aqueduct_executor.operators.utils.utils import time_it

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--spec", required=True)
    args = parser.parse_args()

    spec_json = base64.b64decode(args.spec)
    spec = parse_spec(spec_json)

    time_it(job_name=spec.name, job_type=spec.type.value, step="Running Operator (including IO)")(
        execute.run
    )(spec)



import json
from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd
from aqueduct.utils.serialization import deserialize
from aqueduct_executor.operators.connectors.data import connector, extract, load
from aqueduct_executor.operators.connectors.data.config import S3Config
from aqueduct_executor.operators.connectors.data.s3_serialization import (
    S3UnknownFileFormatException,
    S3UnsupportedArtifactTypeException,
    _s3_deserialization_function_mapping,
    artifact_type_to_s3_serialization_type,
    serialize_val_for_s3,
)
from aqueduct_executor.operators.connectors.data.utils import construct_boto_session
from aqueduct_executor.operators.utils.enums import ArtifactType
from aqueduct_executor.operators.utils.saved_object_delete import SavedObjectDelete
from aqueduct_executor.operators.utils.utils import delete_object
from botocore.client import ClientError


class S3Connector(connector.DataConnector):
    def __init__(self, config: S3Config):
        session = construct_boto_session(config)
        self.s3 = session.resource("s3")
        self.bucket = config.bucket

    def authenticate(self) -> None:
        try:
            # Below is a low-overhead way of checking if the user has access to the bucket.
            # Source: https://stackoverflow.com/a/49817544
            self.s3.meta.client.head_bucket(Bucket=self.bucket)
        except ClientError as e:
            raise Exception(
                "Bucket does not exist or you do not have permission to access the bucket: %s."
                % str(e)
            )

    def discover(self) -> List[str]:
        raise Exception("Discover is not supported for S3.")

    def fetch_object(self, key: str, params: extract.S3Params) -> Any:
        response = self.s3.Object(self.bucket, key).get()
        data = response["Body"].read()

        try:
            s3_serialization_type = artifact_type_to_s3_serialization_type(
                params.artifact_type, params.format
            )

        # Append the S3 filepath to the message for additional error context.
        except S3UnsupportedArtifactTypeException:
            raise S3UnsupportedArtifactTypeException(
                "Unsupported data type %s when fetching file at %s." % (params.artifact_type, key)
            )
        except S3UnknownFileFormatException:
            raise S3UnknownFileFormatException(
                "Unknown S3 file format `%s` for file at path `%s`." % (params.format, key)
            )

        try:
            deserialized_val = deserialize(
                s3_serialization_type,
                params.artifact_type,
                data,
                custom_deserialization_function_mapping=_s3_deserialization_function_mapping,
            )
        except Exception as e:
            print(str(e))
            err_msg = "The file at path `%s` is not a valid %s object." % (
                key,
                params.artifact_type,
            )
            if params.artifact_type == ArtifactType.TABLE and params.format is not None:
                err_msg += " (with S3 file format `%s`)" % params.format.value

            raise Exception(err_msg)

        # Perform some additional type checking after deserialization.
        if params.artifact_type == ArtifactType.STRING:
            if not isinstance(deserialized_val, str):
                raise Exception(
                    "The file at path `%s` is expected to be a string, got %s."
                    % (key, type(deserialized_val))
                )
        elif params.artifact_type == ArtifactType.BOOL:
            if not (isinstance(deserialized_val, bool) or isinstance(deserialized_val, np.bool_)):
                raise Exception(
                    "The file at path `%s` is expected to be a bool, got %s."
                    % (key, type(deserialized_val))
                )
        elif params.artifact_type == ArtifactType.NUMERIC:
            if not (
                isinstance(deserialized_val, int)
                or isinstance(deserialized_val, float)
                or isinstance(deserialized_val, np.number)
            ):
                raise Exception(
                    "The file at path `%s` is expected to be a numeric, got %s."
                    % (key, type(deserialized_val))
                )
        elif params.artifact_type == ArtifactType.DICT:
            if not isinstance(deserialized_val, dict):
                raise Exception(
                    "The file at path `%s` is expected to be a dictionary, got %s."
                    % (key, type(deserialized_val))
                )
        elif params.artifact_type == ArtifactType.TUPLE:
            if not isinstance(deserialized_val, tuple):
                raise Exception(
                    "The file at path `%s` is expected to be a tuple, got %s."
                    % (key, type(deserialized_val))
                )
        elif params.artifact_type == ArtifactType.LIST:
            if not isinstance(deserialized_val, list):
                raise Exception(
                    "The file at path `%s` is expected to be a list, got %s."
                    % (key, type(deserialized_val))
                )

        return deserialized_val

    def extract(self, params: extract.S3Params) -> Any:
        path = json.loads(params.filepath)
        if not isinstance(path, List):
            if len(path) == 0:
                raise Exception("S3 file path cannot be an empty string.")
            if path[-1] == "/":
                # This means the path is a directory, and we will do a prefix search.
                files = []
                for obj in self.s3.Bucket(self.bucket).objects.filter(Prefix=path):
                    # The filter api also returns the directories, so we filter them out.
                    if (obj.key)[-1] != "/":
                        files.append(self.fetch_object(obj.key, params))

                if params.artifact_type == ArtifactType.TABLE and params.merge:
                    # We ignore indexes anyways when serializing the data later, so it's ok to do it earlier here.
                    return pd.concat(files, ignore_index=True)
                else:
                    return tuple(files)
            else:
                # This means the path is a file name, and we do a regular file retrieval.
                return self.fetch_object(path, params)
        else:
            # This means we have a list of file paths.
            files = []
            for key in path:
                if len(key) == 0:
                    raise Exception("S3 file path cannot be an empty string.")
                if key[-1] == "/":
                    raise Exception("Each key in the list must not be a directory, found %s." % key)
                files.append(self.fetch_object(key, params))

            if params.artifact_type == ArtifactType.TABLE and params.merge:
                # We ignore indexes anyways when serializing the data later, so it's ok to do it earlier here.
                return pd.concat(files, ignore_index=True)
            else:
                return tuple(files)

    def load(self, params: load.S3Params, data: Any, artifact_type: ArtifactType) -> None:
        serialization_type = artifact_type_to_s3_serialization_type(artifact_type, params.format)
        serialized_data = serialize_val_for_s3(
            data,
            serialization_type,
            params.format,
        )

        self.s3.Object(self.bucket, params.filepath).put(Body=serialized_data)

    def _delete_object(self, name: str, context: Optional[Dict[str, Any]] = None) -> None:
        """`name` is expected to be the S3 FilePath, found in S3Params."""
        self.s3.Object(self.bucket, name).delete()

    def delete(self, objects: List[str]) -> List[SavedObjectDelete]:
        results = []
        for key in objects:
            results.append(delete_object(key, self._delete_object))
        return results
